/* 
 *   Creation Date: <2000/05/28 02:37:51 samuel>
 *   Time-stamp: <2003/04/29 23:14:22 samuel>
 *   
 *	<osi_sound.c>
 *	
 *	Sound Driver
 *   
 *   Copyright (C) 2000-2003 Samuel Rydh (samuel@ibrium.se)
 *   
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   as published by the Free Software Foundation
 *   
 */

#include "mol_config.h"
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <sys/poll.h>
#include <sys/resource.h>
#ifdef CONFIG_ESOUND
#include <esd.h>
#endif
#include "driver_mgr.h"
#include "debugger.h"
#include "os_interface.h"
#include "verbose.h"
#include "memory.h"
#include "thread.h"
#include "res_manager.h"
#include "osi_driver.h"
#include "pic.h"
#include "async.h"
#include "booter.h"
#include "pci.h"
#include "timer.h"
#include "sound_sh.h"

static struct {
	int	fd;
	int	irq;
	int	async_id;

	int	started;

	/* misc */
	int	rate_limit;		/* user imposed rate restriction */

	/* current mode */
	int	mute;
	int	format;
	uint	rate;			/* in Hz */
	int	bpf;			/* bytes per frame */

	/* sound parameters */
	int	fragsize;		/* client fragsize */
	int	fragticks;		/* mticks for a frag */

	/* ring buffer */
	char	*ringbuf;		/* ringbuf (if non-NULL) */
	int	ringfrags;		/* #frags in ringbuf */

	volatile uint ringind;		/* last frag passed to hardware */
	volatile int stop_ring;
} ss;


/************************************************************************/
/*	/dev/dsp support						*/
/************************************************************************/

static int
dsp_convert_format( int format, int *fmt, int *channels )
{
	static struct {
		int mol_fmt, linux_fmt;
	} fmt_table[] = {
		{ kSoundFormat_U8_BE,	AFMT_U8,	},
		{ kSoundFormat_S16_BE,	AFMT_S16_BE,	},
		{ kSoundFormat_S16_LE,	AFMT_S16_LE,	},
	};
	int f, i;	

	for( f=-1, i=0; i<sizeof(fmt_table)/sizeof(fmt_table[0]); i++ )
		if( format & fmt_table[i].mol_fmt )
			f = fmt_table[i].linux_fmt;
	if( f ==  -1 )
		return -1;

	*fmt = f;
	*channels = (format & kSoundFormat_Stereo)? 2:1;
	return 0;
}

static void
dsp_close( void )
{
	close( ss.fd );
	ss.fd = -1;
}

static int
dsp_open( int format, int rate, int *fragsize )
{
	int size, err, fmt, channels, rate_, fmt_, channels_;
	uint frag = (4 << 16) | 12;	/* 4 x 4K */
	char *name;

	if( is_osx_boot() ) {
		//frag = (4 << 16) | 9;	/* 4 x 512 */
		frag = (4 << 16) | 8;	/* 4 x 256 */
	}

	if( ss.started )
		return 1;

	if( ss.fd != -1 || dsp_convert_format( format, &fmt, &channels ))
		return -1;

	if( !(name=get_filename_res("sound_device")) )
		name = "/dev/dsp";
	if( (ss.fd=open(name, O_WRONLY | O_NONBLOCK)) < 0 )
		return 1;

	err=0;
	fmt_ = fmt; channels_ = channels; rate_ = rate;
	if( ioctl( ss.fd, SNDCTL_DSP_SETFMT, &fmt ) < 0 || fmt_ != fmt )
		err--;

	if( ioctl( ss.fd, SNDCTL_DSP_CHANNELS, &channels ) < 0 || channels_ != channels )
		err--;

	if( ioctl( ss.fd, SNDCTL_DSP_SPEED, &rate ) < 0 || rate_ != rate )
		err--;

	ioctl( ss.fd, SNDCTL_DSP_SETFRAGMENT, &frag );
	ioctl( ss.fd, SNDCTL_DSP_GETBLKSIZE, &size );

	if( (1 << (frag & 0xffff)) != size ) {
		printm("Warning: fragsize not accepted\n");
		size = 32 * 1024;
	}
	if( fragsize )
		*fragsize = size;

	if( err ) {
		close( ss.fd );
		ss.fd = -1;
	}
	return err;
}

/* 0: ok, 1: might be supported, <0: unsupported/error */
static int
dsp_query_sound_mode( int format, int rate, int *fragsize )
{
	int ret;
	if( (ret=dsp_open( format, rate, fragsize )) )
		return ret;
	dsp_close();
	return 0;
}

static void
dsp_flush_sound( void )
{
	if( ioctl( ss.fd, SNDCTL_DSP_POST ) < 0 )
		perrorm("SNDCTL_DSP_POST\n");
}


/************************************************************************/
/*	generic routines						*/
/************************************************************************/

static int
query_mode( int format, int rate )
{
	if( ss.rate_limit && rate > ss.rate_limit )
		return -1;
	return dsp_query_sound_mode( format, rate, NULL );
}

static int
set_mode( int format, int rate )
{
	int ret, fragsize;
	
	if( ss.started )
		return 1;
	if( (ret=dsp_query_sound_mode( format, rate, &fragsize )) )
		return ret;
	/* printm("SoundMode: %x, %d Hz, fragsize %d\n", format, rate, fragsize ); */
	
	ss.rate = rate;
	ss.bpf = (format & kSoundFormat_Len1Mask) ? 1 : 2;
	ss.bpf *= (format & kSoundFormat_Stereo ) ? 2 : 1;
	ss.fragsize = fragsize;
	ss.fragticks = usecs_to_mticks_( fragsize * 1000 ) * 1000 / (ss.bpf  * ss.rate);
	ss.format = format;
	return 0;
}

static void
sndpipe_empty( int fd, int events ) 
{
	/* disable async handler until IRQ is raised... */
	set_async_handler_events( ss.async_id, 0 );
	irq_line_hi( ss.irq );
}

static void
ringbuf_thread( void *dummy )
{
	char *p;

	setpriority( PRIO_PROCESS, getpid(), -20 );
	while( !ss.stop_ring ) {
		ss.ringind++;
		if( ss.ringind == ss.ringfrags ) {
			ss.ringind = 0;
			irq_line_hi( ss.irq );
		} else {
			abort_doze();
		}
		p = ss.ringbuf + (ss.ringind % ss.ringfrags) * ss.fragsize;
		
		write( ss.fd, p, ss.fragsize );
	}
	setpriority( PRIO_PROCESS, getpid(), 0 );
	ss.stop_ring = -1;
}

static void
stop_sound( void )
{
	if( !ss.started )
		return;

	if( ss.ringbuf ) {
		ss.stop_ring = 1;
		while( ss.stop_ring != -1 )
			;
	} else {
		delete_async_handler( ss.async_id );
	}
	dsp_close();
	ss.started = 0;
}

static int
start_sound( void ) 
{
	if( ss.started || !ss.format )
		return -1;

	if( dsp_open( ss.format, ss.rate, NULL ) ) {
		/* use timer fallback? */
		return -1;
	}
	ss.started = 1;

	if( ss.ringbuf ) {
		ss.ringind = -1;
		ss.stop_ring = 0;
		create_thread( (void*)ringbuf_thread, NULL, "audio-thread" );
	} else {
		ss.async_id = add_async_handler( ss.fd, 0, sndpipe_empty, 0 );
	}
	return 0;
}

static void
flush_sound( void )
{
	if( !ss.started )
		return;
	dsp_flush_sound();
}

static void
resume_sound( void )
{
	if( ss.started )
		set_async_handler_events( ss.async_id, POLLOUT );
}

static void
pause_sound( void )
{
	if( ss.started )
		set_async_handler_events( ss.async_id, 0 );
}

static int
query_pipefull( void )
{
	/* unimplemented */
	return -1;
}



/************************************************************************/
/*	OSI interface							*/
/************************************************************************/

static int
osip_sound_irq_ack( int sel, int *params )
{
	return irq_line_low( ss.irq );
}

static int
osip_sound_cntl( int sel, int *params )
{
	switch( params[0] ) {
	case kSoundCheckAPI:
		if( params[1] != OSI_SOUND_API_VERSION ) {
			printm("Incompatible MOLAudio version!\n");
			return -1;
		}
		break;
	case kSoundStart:
		return start_sound();
	case kSoundSetupRing: /* nfrags, buf */
		ss.ringfrags = params[1];
		ss.ringbuf = params[2] ? transl_mphys(params[2]) : NULL;
		break;
	case kSoundGetRingSample: 
		return ss.ringind * ss.fragsize/ss.bpf;
	case kSoundStop:
		stop_sound();
		break;
	case kSoundFlush:
		flush_sound();
		break;
	case kSoundResume:
		resume_sound();
		break;
	case kSoundPause:
		pause_sound();
		break;
	case kSoundGetFragsize:
		return ss.fragsize;		
	case kSoundRouteIRQ:
		oldworld_route_irq( params[1], &ss.irq, "sound" );
		break;
	case kSoundQueryFormat: /* format, rate */
		return query_mode( params[1], params[2] );
	case kSoundSetFormat: /* format, rate */
		return set_mode( params[1], params[2] );
	case kSoundQueryPipeFull: /* pipe full? */
		return query_pipefull();
	case kSoundOSXDriverVersion:
		set_driver_ok_flag( kOSXAudioDriverOK, params[1] == OSX_SOUND_DRIVER_VERSION );
		break;
	case kSoundClassicDriverVersion:
		if( params[1] < CLASSIC_SOUND_DRIVER_VERSION )
			printm("Install the latest MOLAudio extension!\n");
		break;
	default:
		return -1;
	}
	return 0;
}

/* phys_buffer, len, restart */
static int
osip_sound_write( int sel, int *params )
{
	char *buf = transl_mphys( params[0] );
	int ret, len = params[1];

	if( !ss.started || !buf )
		return -1;

	/* pad... buffer is guaranteed to be at least of size ss.fragsize */
	while( len < ss.fragsize )
		buf[len++] = 0;

	if( (ret=write( ss.fd, buf, ss.fragsize )) != ss.fragsize )
		printm("Write error: %d\n", ret );

	if( params[2] )
		set_async_handler_events( ss.async_id, POLLOUT );
	return 0;
}

/* hwVolume, speakerVolume, hwMute */
static int
osip_volume( int sel, int *params )
{
	uint vol, lvol = (params[0] & 0xffff0000) >> 16;
	uint rvol = params[0] & 0xffff;
	ss.mute = params[2] ? 1:0;

	if( ss.fd < 0 )
		return 0;

	if( (lvol=lvol*100/0x100) > 100 )
		lvol = 100;
	if( (rvol=rvol*100/0x100) > 100 )
		rvol = 100;
	vol = rvol*0x100 | lvol;

	/* printm("Volume %d/%d\n", lvol, rvol ); */

	ioctl( ss.fd, SOUND_MIXER_WRITE_VOLUME, &vol );
	return 0;
}

#if 0
static void
play_startboing_entry( void *dummy )
{
	char *name = get_filename_res("startboing_file");
	sigset_t oldmask, set;
	char buf[512], *p;
	int fd, n, n2, ret;
	int format = AFMT_S16_BE;
	int rate = 22050;
	int channels = 2;

	ss.fd = open("/dev/dsp", O_WRONLY );
#if 1
	if( ss.fd < 0 ) {
		printm("no DSP!\n");
		return;
	}
#endif
	if( (fd=open( name, O_RDONLY )) < 0 ) {
		printm("Could not open '%s'", name);
		return;
	}
	ioctl( ss.fd, SNDCTL_DSP_SETFMT, &format );
	ioctl( ss.fd, SNDCTL_DSP_CHANNELS, &channels );
	ioctl( ss.fd, SNDCTL_DSP_SPEED, &rate );

	n = lseek( fd, 0, SEEK_END );
	lseek( fd, 0, SEEK_SET );

	while( n > 0 ) {
		ret = read( fd, buf, sizeof(buf) );
		if( ret < 0 )
			break;
		n -= ret;
		for( p=buf; ret>0; p+=n2, ret-=n2 ) {
			n2 = write( ss.fd, p, ret );
			if( n2 == -1 && (errno != EAGAIN && errno != EINTR) ) {
				printm("X\n");
				break;
			} else if( n2 == -1 ) {
				n2 = 0;
				usleep(1);
			}
		}
	}
//	sigfillset( &set );
//	pthread_sigmask( SIG_BLOCK, &set, &oldmask );
	TEMP_FAILURE_RETRY( ioctl( ss.fd, SNDCTL_DSP_SYNC, NULL ));
//	pthread_sigmask( SIG_SETMASK, &oldmask, NULL );

	printm("SOUND PLAYING DONE\n");

	close( fd );
}
#endif

/************************************************************************/
/*	init / cleanup							*/
/************************************************************************/

static int 
osi_sound_init( void )
{
	static pci_dev_info_t pci_config = { 0x6666, 0x6668, 0x02, 0x0, 0x0000 };
	char *drv = is_oldworld_boot() ? "sound" : NULL;
	int limit;
	
	memset( &ss, 0, sizeof(ss) );
	ss.fd = -1;

	if( !register_osi_driver( drv, "mol-audio", is_classic_boot()? &pci_config : NULL, &ss.irq ) ) {
		printm("Failed to register sound card\n");
		return 0;
	}

	/* handle user rate limits */
	if( (limit=get_numeric_res("max_sound_rate")) > 0 )
		ss.rate_limit = limit;

	/* the osi procs must be registered even if sound is unavailable */
	os_interface_add_proc( OSI_SOUND_CNTL, osip_sound_cntl );
	os_interface_add_proc( OSI_SOUND_IRQ_ACK, osip_sound_irq_ack );
	os_interface_add_proc( OSI_SOUND_WRITE, osip_sound_write );
	os_interface_add_proc( OSI_SOUND_SET_VOLUME, osip_volume );
	return 1;
}

static void
osi_sound_cleanup( void )
{
}

driver_interface_t osi_sound_driver = {
    "osi_sound", osi_sound_init, osi_sound_cleanup
};
