/* 
 *   Creation Date: <1999/03/05 16:14:42 samuel>
 *   Time-stamp: <2003/04/29 23:15:55 samuel>
 *   
 *	<pci-brdiges.c>
 *	
 *	PCI bridges (bandit, chaos and grackle)
 *   
 *   Copyright (C) 1999, 2000, 2001, 2002, 2003 Samuel Rydh (samuel@ibrium.se)
 *   
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   as published by the Free Software Foundation
 *   
 */

#include "mol_config.h"

/* #define VERBOSE */

#include "verbose.h"
#include "promif.h"
#include "ioports.h"
#include "debugger.h"
#include "pci.h"
#include "debugger.h"
#include "extralib.h"
#include "byteorder.h"
#include "driver_mgr.h"

SET_VERBOSE_NAME("pci-bridges")

/* #define BRIDGE_DEBUG */
/* #define DEBUG_ADDR		0x0068	*/


/* this hold for bandit & chaos */
#define BANDIT_CFG_ADDR_OFFS	0x800000
#define BANDIT_CFG_DATA_OFFS	0xc00000

/* This is for Grackle in map B - note that the address is by 0xffffc -
 * i.e. we must make sure we add maps. Motorola suggests FEC00CF8 and
 * FEE00CFE.
 */
#define GRACKLE_CFG_ADDR	0xFEC00000
#define GRACKLE_CFG_DATA	0xFEE00000

#define BANDIT_DEV_FN		(11<<3)
#define GRACKLE_DEV_FN		0

/* Chaos bridge:
 *   It appears as if OF reads data from offset c00004 whenever 
 *   the config address ends at 0x4 or 0xc, probably due to a 64-bit
 *   bus interface or something (the linux kernel doesn't do this).
 */

typedef struct pci_bridge {
	char 	*bridge_name;	/* verbose name */
	ulong	mbase;		/* register base (0xF8000000) */

	ulong	addr_mbase;	/* config register port */
	ulong	data_mbase;	/* config data port */

	int	bridge_type;	/* chaos or bandit */
	void	*type_data;	/* bridge specific data goes here */
	
	int	first_bus;
	int	last_bus;

	pci_addr_t (*interpret_addr)( struct pci_bridge *b, ulong pciaddr, int *offs );

	struct pci_bridge  *next;

	/* emulated hardware registers */
	ulong	addr;		/* address register */
} pci_bridge_t;

static pci_bridge_t *all_bridges;

enum {
	bandit_bridge, chaos_bridge, grackle_bridge
};


/* calculates the inverse of 
 *    
 *      addr = (1UL << (dev_fn >> 3)) + ((dev_fn & 7) << 8) + (offset & ~3)) 
 * 
 * (for bandit & chaos)
 */
static pci_addr_t
bandit_interpret_addr( pci_bridge_t *b, ulong pciaddr_le, int *offs )
{
	ulong tbit, pciaddr_be;
	int val, hit=0, fn_hi=0;
	int dev_fn, bus;
	
	pciaddr_be = ld_le32( &pciaddr_le );
	
	/* [16+5 device num] [3 bits function] [8 bits address offset] */

	*offs = (pciaddr_be & ~3) & 0xff;
	dev_fn = (pciaddr_be >> 8) & 7;
	bus = b->first_bus;	/* XXX: fix me */

	for( val=11, tbit=0x800; tbit; tbit=tbit<<1, val++ ){
		if( pciaddr_be & tbit ) {
			hit++;
			dev_fn |= val<<3;
			fn_hi = val;
		}
	}
	if( hit != 1 )
		return -1;

	return PCIADDR_FROM_BUS_DEVFN( bus, dev_fn );
}

/* Grackle: LE:  offset.B&~0x3 dev_fn.B bus.B 0x80  */
static pci_addr_t
grackle_interpret_addr( pci_bridge_t *b, ulong pciaddr_le, int *offs)
{
	int bus, dev_fn;
	
	*offs = pciaddr_le >> 24;
	dev_fn = (pciaddr_le >> 16) & 0xff;
	bus = (pciaddr_le >> 8) & 0xff;

	if( bus < b->first_bus || bus > b->last_bus )
		return -1;

	/* possibly an error should be returned if the enable bit is NOT set */
	return PCIADDR_FROM_BUS_DEVFN( bus, dev_fn );
}

/************************************************************************/
/*	Bridges								*/
/************************************************************************/

static pci_dev_info_t bandit_pci_config = {
	/* vendor, device ID, revision, prog, class */
	0x106b, 0x0001, 0x03, 0x0/*?*/, 0x0600
};
static pci_dev_info_t grackle_pci_config = {
	/* vendor, device ID, revision, prog, class */
	0x1057, 0x0002, 0x03, 0x0/*?*/, 0x0600
};

static void
init_bandit_bridge( mol_device_node_t *dn, struct pci_bridge *bridge )
{
	pci_addr_t addr = PCIADDR_FROM_BUS_DEVFN( bridge->first_bus, BANDIT_DEV_FN );
	add_pci_device( addr, &bandit_pci_config, NULL );
}

static void
init_chaos_bridge( mol_device_node_t *dn, struct pci_bridge *bridge )
{
}

static void
init_grackle_bridge( mol_device_node_t *dn, struct pci_bridge *bridge )
{
	pci_addr_t addr = PCIADDR_FROM_BUS_DEVFN( bridge->first_bus, GRACKLE_DEV_FN );
	add_pci_device( addr, &grackle_pci_config, NULL );
}


/************************************************************************/
/*	I / O								*/
/************************************************************************/

static ulong 
reg_io_read( ulong addr, int len, void *u_bridge )
{
	pci_bridge_t *bridge = (pci_bridge_t*)u_bridge;
	int offs;

	/* address port */
	if( (addr & ~0x3) == bridge->addr_mbase ) {
		ulong buf[2];
		buf[0] = bridge->addr;
		return read_mem( (char*)buf+(addr&3), len );
	}
	/* data port */
	if( (addr & ~0x7) == bridge->data_mbase ) {
		pci_addr_t pci_addr = bridge->interpret_addr( bridge, bridge->addr, &offs );
		ulong ret = (addr < 0) ? 0 : read_pci_config( pci_addr, offs+(addr&3), len );
#if defined(DEBUG_ADDR) 
		if( addr == DEBUG_ADDR )
			printm("PCI, read [%d] (%04x)+%02X: %08lX\n", len, pci_addr, offs, ret );
#endif
		return ret;
	}
	return 0;
}

static void 
reg_io_write( ulong addr, ulong data, int len, void *u_bridge )
{
	pci_bridge_t *bridge = (pci_bridge_t*)u_bridge;
	int offs;
	
	/* Address port */
	if( (addr & ~0x3) == bridge->addr_mbase ) {
		ulong buf[2];
		buf[0] = bridge->addr;
		write_mem( (char*)buf+(addr&3), data, len );
		bridge->addr = buf[0];
		return;
	}

	/* Data port - note that Chaos uses a 64-bit bus interface (or?),
	 * placing the data in the upper word for xxx4/xxxc accesses 
	 */
	if( (addr & ~0x7) == bridge->data_mbase ) {
		pci_addr_t pci_addr = bridge->interpret_addr( bridge, bridge->addr, &offs );
		if( pci_addr != -1 )
			write_pci_config( pci_addr, offs+(addr&3), data, len );
#if defined(DEBUG_ADDR) 
		if( addr == DEBUG_ADDR )
			printm("PCI, write [%d] (%04x)+%02X: %08lX\n", len, pci_addr, offs, ret );
#endif
	}
}

static void 
reg_io_print( int isread, ulong addr, ulong data, int len, void *u_bridge )
{
	pci_bridge_t *bridge = (pci_bridge_t*)u_bridge;
	pci_addr_t pci_addr;
	ulong d = data;
	int offs;
	
	if( (addr & ~0x3) == bridge->addr_mbase ) {
#if 0
		pci_addr = bridge->interpret_addr( bridge, bridge->addr, &offs );
		printm("Bridge '%s' %s %08lX ; (%04x+%02x)\n", bridge->bridge_name,
		       isread ? "read: " : "write:", ld_le32(&data), pci_addr, offs );
#endif
		return;
	}

	if( (addr & ~0x7) == bridge->data_mbase ) {
		pci_addr = bridge->interpret_addr( bridge, bridge->addr, &offs );
		addr &= ~0x4;

		if( len == 4 )
			d = ld_le32( &data );
		else if( len == 2 )
			d = ld_le16( (short*)&data +1 );
		printm("Bridge '%s-%d' %s (%04x+%02lX): %08lX\n", bridge->bridge_name,
		       len, isread? "read " : "write", pci_addr, offs+(addr&3), d);
		return;
	}

	printm("bridge ???? %08lX: (len=%d) %s %08lX\n", addr, len,
	       isread ? "read: " : "write:", data );
	/* stop_emulation(); */
}


/************************************************************************/
/*	init / cleanup							*/
/************************************************************************/

static io_ops_t ops = {
	.read	= reg_io_read,
	.write	= reg_io_write, 
	.print	= reg_io_print
};

static void
init_bridges( char *device_name, int bridge_type )
{
	mol_device_node_t *dn = prom_find_devices( device_name );
	pci_bridge_t *bridge;
	char namebuf[40];
	int first_bus, last_bus;

	for( ; dn ; dn=dn->next ) {
		if( dn->n_addrs != 1 ) {
			printm("%s: Expecting 1 addrs. Got %d)\n", device_name, dn->n_addrs );
			continue;
		}

		if( pci_get_bus_number(dn, &first_bus, &last_bus) ) {
			LOG("Couldn't get bus number\n");
			continue;
		}
		bridge = calloc( sizeof(struct pci_bridge), 1 );
		bridge->bridge_name = strdup( device_name );
		bridge->next = all_bridges;
		all_bridges = bridge;
		
		bridge->mbase = dn->addrs[0].address;
		bridge->first_bus = first_bus;
		bridge->last_bus = last_bus;
		bridge->bridge_type = bridge_type;

		/* is this a Grackle ? */
		if( !strcmp(device_name, "pci") ) {
			bridge->addr_mbase = GRACKLE_CFG_ADDR;
			bridge->data_mbase = GRACKLE_CFG_DATA;
			bridge->interpret_addr = grackle_interpret_addr;
		} else {
			/* this is a bandit or chaos */
			bridge->addr_mbase = bridge->mbase + BANDIT_CFG_ADDR_OFFS;
			bridge->data_mbase = bridge->mbase + BANDIT_CFG_DATA_OFFS;
			bridge->interpret_addr = bandit_interpret_addr;
		}
		
		/* the minimum IO-area size is 16 bytes, we need only 4 (8 for chaos) bytes though */
		snprintf( namebuf, 40, "%s-%d-%d-addr", device_name, first_bus, last_bus );
		add_io_range( bridge->addr_mbase, 0x10, namebuf, 0, &ops, (void*)bridge );
		snprintf( namebuf, 40, "%s-%d-%d-data", device_name, first_bus, last_bus );
		add_io_range( bridge->data_mbase, 0x10, namebuf, 0, &ops, (void*)bridge );

		switch( bridge_type ) {
		case chaos_bridge:
			init_chaos_bridge( dn, bridge );
			break;
		case bandit_bridge:
			init_bandit_bridge( dn, bridge );
			break;
		case grackle_bridge:
			init_grackle_bridge( dn, bridge );
			break;
		default:
			LOG("Warning: Unsupported bridge type\n");
			break;
		}
		
		VPRINT("PCI-bridge '%s' (bus %d..%d) installed at 0x%08lx\n", 
		       device_name, first_bus, last_bus, bridge->mbase );
	}
}

static int
pci_bridges_init( void )
{
	all_bridges = NULL;
	init_bridges( "bandit", bandit_bridge );
	init_bridges( "chaos", chaos_bridge );
	init_bridges( "pci", grackle_bridge );
	return 1;
}

static void
pci_bridges_cleanup( void )
{
	pci_bridge_t *br, *next;
	for( br=all_bridges; br; br=next ){
		next = br->next;
		if( br->bridge_name )
			free( br->bridge_name );
		if( br->type_data )
			free( br->type_data );
		free( br );
	}
	all_bridges = NULL;
}

driver_interface_t pci_bridges_driver = {
	"pci-bridges", pci_bridges_init, pci_bridges_cleanup
};
