/* 
 *   Creation Date: <97/05/26 02:10:43 samuel>
 *   Time-stamp: <2003/03/29 15:15:10 samuel>
 *   
 *	<misc.c>
 *	
 *	Kernel interface
 *   
 *   Copyright (C) 1997-2003 Samuel Rydh (samuel@ibrium.se)
 *   
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   as published by the Free Software Foundation
 *   
 */

#include "archinclude.h"

#include <linux/vmalloc.h>
#include <asm/uaccess.h>
#include <asm/prom.h>

#include "multiplexer.h"
#include "mac_registers.h"
#include "kernel_vars.h"
#include "misc.h"
#include "mmu.h"
#include "asmfuncs.h"
#include "emu.h"
#include "prom.h"
#include "version.h"
#include "mtable.h"
#include "performance.h"
#include "constants.h"

kernel_vars_t *
alloc_kvar_pages( void )
{
	kernel_vars_t *kv;
	char *ptr;
	int i, order=0;

	for( i=1, order=0; i<NUM_KVARS_PAGES; i=i<<1, order++ )
		;
	if( !(kv = (kernel_vars_t*)__get_free_pages( GFP_KERNEL, order )) )
		return NULL;

	/* To be able to export the kernel variables to user space, we
	 * must set the PG_reserved bit. This is due to a check
	 * in remap_pte_range() in kernel/memory.c (is this bug or a feature?).
	 */
	for(ptr=(char*)kv, i=0; i<NUM_KVARS_PAGES; i++, ptr+=0x1000 )
		compat_set_page_reserved( ptr );

	return kv;
}

void
free_kvar_pages( kernel_vars_t *kv )
{
	char *ptr = (char*)kv;
	int i, order;

	for( i=0; i<NUM_KVARS_PAGES; i++, ptr+=0x1000 )
		compat_clear_page_reserved( ptr );

	for( i=1, order=0; i<NUM_KVARS_PAGES; i=i<<1, order++ )
		;
	free_pages( (ulong)kv, order );
}


/************************************************************************/
/*	copy kernel OF tree						*/
/************************************************************************/

static struct device_node *
get_prom_root( void ) 
{
	struct device_node *root;
	
	root = find_path_device("/");
	return root;
}

static void *
prom_copy_( char *buf, void *src, int size, int *total_size ) 
{
	char *ret = NULL;

	/* Pad to word boundary */
	if( (*total_size & 3) )
		*total_size += 4 - (*total_size & 3);

	if( buf ) {
		if( !src )
			memset( buf + *total_size, 0, size );
		else
			memcpy( buf + *total_size, src, size );
		ret = buf + *total_size;
	}
	*total_size += size;
	return ret;
}

static int 
do_prom_copy_node( char *ptr, struct device_node *src )
{
	mol_device_node_t *dest, dummy_dest;
	p_property_t *prev;
	struct property *pr;
	struct device_node *dn;
	int tsize=0;
	
	if( !(dn=get_prom_root()) )
		return -ENOENT;
	if( !src )
		src = dn;
	else {
		for( ; dn && dn!=src ; dn=dn->allnext )
			;
		if( !dn )
			return -EFAULT;
	}
	if( !(dest=prom_copy_( ptr, NULL, sizeof(mol_device_node_t), &tsize)))
		dest = &dummy_dest;
	dest->node = src->node;
	dest->n_addrs = src->n_addrs;
	dest->n_intrs = src->n_intrs;

	/* copy kernel node references */
	dest->parent = (mol_device_node_t*) src->parent;
	dest->child = (mol_device_node_t*) src->child;
	dest->sibling = (mol_device_node_t*) src->sibling;
	dest->next = (mol_device_node_t*) src->next;
	dest->allnext = (mol_device_node_t*) src->allnext;

	/* arrays */
	dest->addrs = prom_copy_( ptr, src->addrs, sizeof(struct address_range)*src->n_addrs, &tsize );
	dest->intrs = prom_copy_( ptr, src->intrs, sizeof(struct interrupt_info)*src->n_intrs, &tsize );

	/* strings */
	if( src->name )
		dest->name = prom_copy_( ptr, src->name, strlen(src->name)+1, &tsize );
	if( src->type )
		dest->type = prom_copy_( ptr, src->type, strlen(src->type)+1, &tsize );
	if( src->full_name )
		dest->full_name = prom_copy_( ptr, src->full_name, strlen(src->full_name)+1, &tsize );
	
	/* properties */
	prev = 0;
	for( pr=src->properties; pr; pr=pr->next ) {
		p_property_t dummy, *t;
		if( !(t = prom_copy_( ptr, NULL, sizeof(p_property_t), &tsize )) )
			t = &dummy;
		t->length = pr->length;

		if( pr->name )
			t->name = prom_copy_( ptr, pr->name, strlen(pr->name)+1, &tsize );
		t->value = prom_copy_( ptr, pr->value, pr->length, &tsize );
		t->next = prev;
		prev = t;
	}
	if( prev )
		dest->properties = prev;

	if( tsize & 4 )
		tsize += 4 - (tsize & 3);

	return tsize;
}

static int
prom_copy_node( char *dest, struct device_node *dn, int bufsize )
{
	int ret;
	
	if( !dest )
		return do_prom_copy_node( NULL, dn );

	if( verify_area(VERIFY_WRITE, dest, bufsize) )
		return -EFAULT;

	if( (ret=do_prom_copy_node(NULL, dn)) > bufsize || ret < 0 )
		return -EFAULT;

	do_prom_copy_node( dest, dn );
	return 0;
}


/************************************************************************/
/*	SMP IPIs							*/
/************************************************************************/
		
#ifdef CONFIG_SMP
static void
dummy_ipi( void *dummy )
{
	/* we don't need to _do_ anything, the exception itself is sufficient */
}
#endif

static inline
void send_ipi( void )
{
#ifdef CONFIG_SMP
	smp_call_function( dummy_ipi, NULL, 1, 0 );
#endif
}


/************************************************************************/
/*	the kernel module entrypoint					*/
/************************************************************************/

static inline void
bench_entry( kernel_vars_t *kv )
{
	ulong now;
	asm( "mftb %0" : "=r" (now) : );

}

static inline void
bench_exit( kernel_vars_t *kv )
{
	ulong now;
	asm( "mftb %0" : "=r" (now) : );
}

ulong 
sys_macos_multiplexer( int what, int sess_id, ulong arg1, ulong arg2, ulong arg3, ulong arg4 )
{
	kernel_vars_t *kv;
	ulong ret, r;

	if( !suser() )
		return -EPERM;	

	// This selector is used by the kernel-assembly code: r4-r6=parameters, r7=function, r8=kvars
	if( what == fCallKernelFunc ) {
		typedef int (*kernelfunc_t)( kernel_vars_t *, ulong, ulong, ulong );
		kv = (kernel_vars_t*)arg4;
		TICK_CNTR_PUSH( kv );
		ret = (*(kernelfunc_t)arg3)( kv, (ulong)sess_id, arg1, arg2 );
		TICK_CNTR_POP( kv, in_kernel );
		return ret;
	}

	if( (sess_id < 0 || sess_id >= MAX_NUM_SESSIONS) && what != fInit ) {
		printk("Bad session number %d\n", sess_id );
		return 1;
	}
	if( !(kv=g_sesstab->kvars[sess_id] ) ) {
		if( what != fInit && what != fGetMolModVersion && what != fGetPVR ) {
			printk("Session %d not initialized (selector %d)\n", sess_id, what );
			return 1;
		}
	}

	ret = 0;
	switch( what ) {
	case fInit: /* arg1 = (new) session_magic */
		ret = initialize_session( sess_id, arg1 );
		break;
	case fCleanup:
		destroy_session( sess_id );
		break;

	case fBreakpointFlags:
		kv->break_flags = arg1;
		kv->mregs.flag_bits &= ~fb_DbgSingleStep;
		kv->mregs.flag_bits |= (arg1 & BREAK_SINGLE_STEP)? fb_DbgSingleStep : 0;
		msr_altered(kv);	/* updates _msr (sets MSR_SE if needed) */
		break;

	case fGetMregs:
		/* ret = kernel virtual address of mregs */
		ret = (ulong)&kv->mregs;
		break;
	case fGetMregsPhys:
		ret = virt_to_phys( &kv->mregs );
		break;

	case fGetMolModVersion:
		ret = MOL_VERSION;
		break;

	case fSprChanged:
		msr_altered(kv);
		mmu_altered(kv);
		break;

	case fMsrChanged:
		msr_altered(kv);
		break;

	case fSetMacRam:
		/* void ( char *lvbase, size_t size, ulong mbase ) */
		kv->mmu.linux_ram_base = (char*)arg1;
		kv->mmu.ram_size = arg2;
		kv->mmu.mac_ram_base = arg3;
		mtable_tune_alloc_limit( kv, arg2/(1024 * 1024) );
		break;
	
	case fAddIORange:
       		/* void ( ulong mbase, int size, void *usr_data )*/
		add_io_trans( kv, arg1, arg2, (void*)arg3 );
		break;
	case fRemoveIORange:
		/* void ( ulong mbase, int size ) */
		remove_io_trans( kv, arg1, arg2 );
		break;

	case fPromGetNodeSize: /* arg1 = node_ptr, 0=root */
		ret = prom_copy_node( NULL, (struct device_node*)arg1, 0 );
		break;

	case fPromCopyNode: /* arg1 = dest, arg2 = kernel_node, arg3 = bufsize */
		ret = prom_copy_node( (char*)arg1, (struct device_node*)arg2, arg3 );
		break;

	/* this selector is going to die peacefully */
	case fKernelMemmove: /* arg1 = dest, arg2 = src, arg3 = size */
		memmove( (char*)arg1, (char*)arg2, arg3 );		
		break;

	case fIdleReclaimMemory:
		mtable_reclaim( kv );
		break;

	case fSMPSendIPI:
		send_ipi();
		break;

	case fEmulateTLBIE:  /* arg1 = pageindex (bits 4-19) */
		//do_tlbie( kv, arg1 );
		flush_ea_range( kv, (arg1 & ~0xf0000000), 0x1000 );
		break;

	case fEmulateTLBIA:  /* for debugging only */
		clear_all_vsids( kv );
		break;
	
	case fAddMMUMapping: /* arg1 = struct mmu_mapping *m */
		mmu_add_map( kv, (struct mmu_mapping*) arg1 );
		break;
	case fRemoveMMUMapping: /* arg1 = struct mmu_mapping *m */
		mmu_remove_map( kv, (struct mmu_mapping*) arg1 );
		break;

	case fGetKernelVersion:
		ret = linux_vers;
		break;

	case fAllocEmuAccelSlot: /* EMULATE_xxx, param, ret_addr -- ind */
		ret = alloc_emuaccel_slot( kv, arg1, arg2, arg3 );
		break;
		
	case fMapinEmuAccelPage: /* arg1 = mphys */
		ret = mapin_emuaccel_page( kv, arg1 );
		break;

	/* THE FOLLOWING SELECTORS SHOULD NOT BE USED IN THE EMULATION */
	case fGetKvarBase: /* for debugging only */
		ret = (ulong)kv;
		break;		

	case fGetPTE: /* arg1 = segment_ident, arg2 = address, arg3 = PTE*  */
		ret = dbg_get_PTE( kv, arg1, arg2, (mPTE_t*)arg3 );
		break;
		
	case fGetPhysicalAddr: /* arg1 = lvptr, arg2 = ret_pageinfo */
		ret = dbg_get_linux_page( arg1, (struct linux_page*)arg2 );
		break;

	case fTranslateIEa: /* context, lvptr, mphys_ret* */
		ret = dbg_translate_ea( kv, arg1, arg2, (ulong*)arg3, 0 );
		break;

	case fTranslateDEa: /* context, lvptr, mphys_ret* */
		ret = dbg_translate_ea( kv, arg1, arg2, (ulong*)arg3, 1 );
		break;
	
	case fSetupFBAccel:
		setup_fb_acceleration( kv, (char*)arg1, arg2, arg3 );
		break;
	case fGetDirtyFBLines:
		ret = get_dirty_fb_lines( kv, (short*)arg1, arg2 );
		break;
	case fGetPVR: /* retpvr */
		if( verify_area( VERIFY_WRITE, (char*)arg1, sizeof(ulong) ))
			ret = -EFAULT;
		else {
			asm volatile("mfpvr %0" : "=r" (r) : );
			*(ulong*)arg1 = r;
		}
		break;

	case fTrackDirtyRAM:
		ret = track_lvrange( kv, (char*)arg1, (size_t)arg2 );
		break;
	case fGetDirtyRAM:
		ret = get_track_buffer( kv, (char*)arg1 );
		break;
	case fSetDirtyRAM:
		set_track_buffer( kv, (char*)arg1 );
		break;
	case fGetPerformanceInfo:
		ret = get_performance_info( kv, arg1, (char*)arg2, arg3, (int*)arg4 );
		break;
	case fClearPerformanceInfo:
		clear_performance_info( kv );
		break;
	case fGetTBFrequency:
		ret = get_tb_frequency();	/* returns 0 on 2.2 */
		break;
	default:
		printk("MAC: multiplexer: index not implemented. Kernel up to date?\n");
		ret = -ENOSYS;
	}
	return ret;
}
